#include <ncbi.h>
#include <blast/blast.h>

static void LIBCALL count_forward PROTO((BLAST_HSPPtr PNTR, size_t));
static void LIBCALL count_reverse PROTO((BLAST_HSPPtr PNTR, size_t));
static int LIBCALL isforward PROTO((BLAST_HSPPtr,BLAST_HSPPtr));
static int LIBCALL isreverse PROTO((BLAST_HSPPtr,BLAST_HSPPtr));

static int LIBCALL fwdcmp PROTO((BLAST_HSPPtr PNTR, BLAST_HSPPtr PNTR));
static int LIBCALL revcmp PROTO((BLAST_HSPPtr PNTR, BLAST_HSPPtr PNTR));

/* consistn -- determine the consistent "n" parameter for each HSP */
/*
	consistn_dynamic computes consistent "n" individually for each HSP,
	requiring O(N^2) operations worst-case per HSP or O(N^3) operations
	worst-case for the entire list of HSPs, where N is the total number
	of HSPs in the HitList.

	To save some time, for each HSP, two lists of the other HSPs are
	considered:  those HSPs that fall in the upper left rectangle
	above the current HSP and those that fall to the lower right rectangle
	below the current HSP.  In looking at the upper left, the terminal end
	points are considered as to whether they can be linked to the beginning
	of the HSP currently being computed.  In looking to the lower right,
	the starting points of the HSPs are considered as to whether they can
	be reached from the end of the current HSP.

	And of course, for each HSP, only those HSPs that are at least
	as informative as the given HSP are considered, for the purpose
	of using the consistent "n" value in Poisson statistics.
*/
void
consistn_dynamic(bcp, hp0, fraction)
	BLAST_ConfigPtr	bcp; /* unused */
	BLAST_HSPPtr	hp0;
	double		fraction;
{
	BLAST_HSPPtr	hspstack[200];
	BLAST_HSPPtr	PNTR flist0, PNTR rlist0;
	register BLAST_HSPPtr	hp;
	register BLAST_HSPPtr	PNTR flist, PNTR flistmax;
	register BLAST_HSPPtr	PNTR rlist, PNTR rlistmax;
	size_t	i, cnt, cnts[9];
	register int	qsign, ssign, sgn;

	if (hp0 == NULL)
		return;
	if (hp0->next == NULL) {
		hp0->n = 1;
		hp0->fwdptr = hp0->revptr = NULL;
		return;
	}

	/*
	There is one element in cnts[] for every strand combination.
	{-1, 0, +1} X {-1, 0, +1} = 9 combinations
	*/
	cnts[0] = cnts[1] = cnts[2] = cnts[3] = cnts[4] = cnts[5] =
		cnts[6] = cnts[7] = cnts[8] = 0;

	for (hp = hp0; hp != NULL; hp = hp->next) {
		qsign = SIGN(hp->q_seg.frame);
		ssign = SIGN(hp->s_seg.frame);
		sgn = (qsign + 1) * 3 + ssign + 1;
		++cnts[sgn];
		hp->n = 1;
		hp->fwdptr = hp->revptr = NULL;
	}

	for (sgn = 0; sgn < DIM(cnts); ++sgn) {
		if ((cnt = cnts[sgn]) < 2)
			continue;
		if (cnt <= DIM(hspstack)/2)
			flist0 = hspstack;
		else
			flist0 = (BLAST_HSPPtr PNTR) BlastMalloc(sizeof(*flist0) * 2 * cnt);
		/* piggy-back rlist0 at the end of flist0 */
		rlist0 = flistmax = flist0 + cnt;
		rlistmax = rlist0 + cnt;
		/* Convert to linear lists for sorting */
		flist = flist0;
		rlist = rlist0;
		for (hp = hp0; flist < flistmax; hp = hp->next) {
			qsign = SIGN(hp->q_seg.frame);
			ssign = SIGN(hp->s_seg.frame);
			if (sgn != (qsign + 1) * 3 + ssign + 1)
				continue;
			*flist++ = hp;
			*rlist++ = hp;

			hp->xscore = BLAST_INFO_BASIC(hp->kbp, hp->score);
			hp->cd.xscore = hp->xscore;
			hp->cd.qbeg = hp->q_seg.offset + fraction * hp->len;
			hp->cd.sbeg = hp->s_seg.offset + fraction * hp->len;
			hp->cd.qend = hp->q_seg.offset + (1.-fraction) * hp->len;
			hp->cd.send = hp->s_seg.offset + (1.-fraction) * hp->len;
		}

		qsort((CharPtr)flist0, cnt, sizeof(*flist0), (int (*)())fwdcmp);
		qsort((CharPtr)rlist0, cnt, sizeof(*rlist0), (int (*)())revcmp);

		for (rlist = rlist0; rlist < rlistmax; ++rlist)
			(*rlist)->user.data.ptrvalue = rlist;
		for (i = 0; i < cnt; ++i) {
			hp = flist0[i];
			count_forward(flist0, i+1);
			hp->n += hp->cd.n;
			rlist = hp->user.data.ptrvalue;
			count_reverse(rlist, rlistmax - rlist);
			hp->n += hp->cd.n;
		}

		if (flist0 != hspstack)
			BlastFree(flist0);
	}
}

static int LIBCALL
fwdcmp(hpp1, hpp2)
	BLAST_HSPPtr	PNTR hpp1, PNTR hpp2;
{
	register BLAST_HSPPtr	hp1 = *hpp1, hp2 = *hpp2;
	register unsigned long	pos1, pos2;

	pos1 = hp1->cd.qbeg;
	pos2 = hp2->cd.qbeg;
	if (pos1 > pos2)
		return 1;
	if (pos1 < pos2)
		return -1;

	pos1 = hp1->cd.sbeg;
	pos2 = hp2->cd.sbeg;
	if (pos1 > pos2)
		return 1;
	if (pos1 < pos2)
		return -1;
	return 0;
}

static int LIBCALL
revcmp(hpp1, hpp2)
	BLAST_HSPPtr	PNTR hpp1, PNTR hpp2;
{
	register BLAST_HSPPtr	hp1 = *hpp1, hp2 = *hpp2;
	register unsigned long	pos1, pos2;

	pos1 = hp1->cd.qend;
	pos2 = hp2->cd.qend;
	if (pos1 < pos2)
		return -1;
	if (pos1 > pos2)
		return 1;

	pos1 = hp1->cd.send;
	pos2 = hp2->cd.send;
	if (pos1 < pos2)
		return -1;
	if (pos1 > pos2)
		return 1;
	return 0;
}

static int
isforward(hp1, hp2)
	register BLAST_HSPPtr	hp1, hp2;
{
	return (hp1->cd.qend < hp2->cd.qbeg && hp1->cd.send < hp2->cd.sbeg);
}


static void LIBCALL
count_forward(list, nelem)
	BLAST_HSPPtr	PNTR list;
	size_t	nelem;
{
	register BLAST_HSPPtr	hp1, hp2;
	register unsigned long	n1plus1;
	register double		xscorelimit;
	register int		i, j;

	for (i = 0; i < nelem; ++i) {
		list[i]->cd.n = 0;
	}
	xscorelimit = list[i-1]->cd.xscore;

	for (i = 0; i < nelem - 1; ++i) {
		hp1 = list[i];
		if (hp1->cd.xscore < xscorelimit)
			continue;
		n1plus1 = hp1->cd.n + 1;
		for (j = i + 1; j < nelem; ++j) {
			hp2 = list[j];
			if (hp2->cd.xscore < xscorelimit)
				continue;
			if (!isforward(hp1, hp2))
				continue;
			hp2->cd.n = MAX(hp2->cd.n, n1plus1);
		}
	}
}

static int
isreverse(hp1, hp2)
	register BLAST_HSPPtr	hp1, hp2;
{
	return (hp1->cd.qbeg > hp2->cd.qend && hp1->cd.sbeg > hp2->cd.send);
}

static void LIBCALL
count_reverse(list, nelem)
	BLAST_HSPPtr	PNTR list;
	size_t	nelem;
{
	register BLAST_HSPPtr	hp1, hp2;
	register unsigned long	n1plus1;
	register double		xscorelimit;
	register int		i, j;

	for (i = 0; i < nelem; ++i)
		list[i]->cd.n = 0;
	xscorelimit = (*list)->cd.xscore;

	for (i = nelem - 1; i > 0; --i) {
		hp1 = list[i];
		if (hp1->cd.xscore < xscorelimit)
			continue;
		n1plus1 = hp1->cd.n + 1;
		for (j = i - 1; j >= 0; --j) {
			hp2 = list[j];
			if (hp2->cd.xscore < xscorelimit)
				continue;
			if (!isreverse(hp1, hp2))
				continue;
			hp2->cd.n = MAX(hp2->cd.n, n1plus1);
		}
	}
}


static int LIBCALL s_pos_less PROTO((BLAST_HSPPtr PNTR,size_t, unsigned long, double));
static int LIBCALL s_pos_greater PROTO((BLAST_HSPPtr PNTR,size_t, unsigned long, double));

static int LIBCALL consistcmp PROTO((BLAST_HSPPtr PNTR, BLAST_HSPPtr PNTR));

/* consistn -- determine the consistent "n" parameter for each HSP */
void
consistn_midpt(hp0)
	BLAST_HSPPtr	hp0;
{
	BLAST_HSPPtr	hspstack[100];
	BLAST_HSPPtr	PNTR list0;
	register BLAST_HSPPtr	hp;
	register BLAST_HSPPtr	PNTR list, PNTR listmax;
	size_t	cnt, cnts[9];
	register int	qsign, ssign, sgn;
	unsigned long	limit;

	if (hp0 == NULL)
		return;
	if (hp0->next == NULL) {
		hp0->n = 1;
		return;
	}

	/*
	There is one element in cnts[] for every strand combination.
	{-1, 0, +1} X {-1, 0, +1} = 9 combinations
	*/
	cnts[0] = cnts[1] = cnts[2] = cnts[3] = cnts[4] = cnts[5] =
		cnts[6] = cnts[7] = cnts[8] = 0;

	for (hp = hp0; hp != NULL; hp = hp->next) {
		qsign = SIGN(hp->q_seg.frame);
		ssign = SIGN(hp->s_seg.frame);
		sgn = (qsign + 1) * 3 + ssign + 1;
		++cnts[sgn];
		hp->n = 1;
		hp->xscore = BLAST_INFO_BASIC(hp->kbp, hp->score);
	}

	for (sgn = 0; sgn < DIM(cnts); ++sgn) {
		if ((cnt = cnts[sgn]) < 2)
			continue;
		if (cnt <= DIM(hspstack))
			list0 = hspstack;
		else
			list0 = (BLAST_HSPPtr PNTR) BlastMalloc(sizeof(*list0) * cnt);
		listmax = list0 + cnt;
		/* Convert to a linear list for sorting by qsort() */
		for (hp = hp0, list = list0; list < listmax; hp = hp->next) {
			qsign = SIGN(hp->q_seg.frame);
			ssign = SIGN(hp->s_seg.frame);
			if (sgn != (qsign + 1) * 3 + ssign + 1)
				continue;
			*list++ = hp;
		}

		qsort((CharPtr)list0, cnt, sizeof(*list0), (int (*)())consistcmp);

		for (list = list0; list < listmax; ++list) {
			hp = *list;
			limit = 2 * hp->s_seg.offset + hp->len;
			hp->n =
				/* count the number that come before the current HSP */
				s_pos_less(list0, list-list0, limit, hp->xscore)
				/* count the number that come after the current HSP */
				+ s_pos_greater(list+1, listmax-list-1, limit, hp->xscore)
				/* count ourselves */
				+ 1;
		}

		if (list0 != hspstack)
			BlastFree(list0);
	}
}

static int LIBCALL
consistcmp(hpp1, hpp2)
	BLAST_HSPPtr	PNTR hpp1, PNTR hpp2;
{
	register BLAST_HSPPtr	hp1 = *hpp1, hp2 = *hpp2;
	register unsigned long	pos1, pos2;

	pos1 = 2 * hp1->q_seg.offset + hp1->len;
	pos2 = 2 * hp2->q_seg.offset + hp2->len;
	if (pos1 > pos2)
		return 1;
	if (pos1 < pos2)
		return -1;

	pos1 = 2 * hp1->s_seg.offset + hp1->len;
	pos2 = 2 * hp2->s_seg.offset + hp2->len;
	if (pos1 > pos2)
		return 1;
	if (pos1 < pos2)
		return -1;
	return 0;
}


/* cnt_hsp_pos -- return the longest increasing subsequence of s_pos values */
static int
s_pos_less(list, len, poslimit, xscorelimit)
	BLAST_HSPPtr	PNTR list;
	size_t	len;
	unsigned long	poslimit;
	double		xscorelimit;
{
	unsigned long	PNTR c, cstack[100];
	BLAST_HSPPtr	PNTR listmax;
	BLAST_HSPPtr	PNTR y, yhp;
	unsigned long	value;
	int		i, j;

	listmax = list + len;

	if (len <= DIM(cstack))
		c = cstack;
	else
		c = (unsigned long PNTR) BlastMalloc(len * sizeof(*c));

	c[0] = 0;
	j = 0;
	for (y = list; y < listmax; ++y) {
		yhp = *y;
		value = 2 * yhp->s_seg.offset + yhp->len;
		if (value >= poslimit)
			continue;
		if (yhp->xscore < xscorelimit)
				continue;
		for (i = j; c[i] > value; --i)
			;
		if (i == j)
			++j;
		c[i+1] = value;
	}
	/* clean-up */
	if (c == cstack)
		return j;
	BlastFree(c);
	return j;
}

/* s_pos_greater -- return the longest increasing subsequence of s_pos values */
static int
s_pos_greater(list, len, poslimit, xscorelimit)
	BLAST_HSPPtr	PNTR list;
	size_t	len;
	unsigned long	poslimit;
	double		xscorelimit;
{
	BLAST_HSPPtr	PNTR y, yhp;
	BLAST_HSPPtr	PNTR listmax;
	unsigned long	PNTR c, cstack[100];
	unsigned long	value;
	int		i, j;

	listmax = list + len;

	if (len <= DIM(cstack))
		c = cstack;
	else
		c = (unsigned long PNTR) BlastMalloc(len * sizeof(*c));

	c[0] = 0;
	j = 0;
	for (y = list; y < listmax; ++y) {
		yhp = *y;
		value = 2 * yhp->s_seg.offset + yhp->len;
		if (value <= poslimit)
			continue;
		if (yhp->xscore < xscorelimit)
			continue;
		for (i = j; c[i] > value; --i)
			;
		if (i == j)
			++j;
		c[i+1] = value;
	}
	/* clean-up */
	if (c == cstack)
		return j;
	BlastFree(c);
	return j;
}
