/* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================*/
#include <ncbi.h>
#include <blast/blast.h>

/* *** Functions which use a global pool are not multiprocessing safe! *** */
/* Programs are required to set up their own per-thread pools */

BLAST_HSPPtr LIBCALL
BlastHSPSave(hlp, hp)
	BLAST_HitListPtr	hlp;
	BLAST_HSPPtr	hp;
{
	BLAST_HSPPtr	hpnew;
 
	hpnew = BlastHSPGet(hlp);
	if (hpnew == NULL)
		return NULL;
	*hpnew = *hp;
	hpnew->hlp = hlp;
	hpnew->next = *hlp->hpp;
	*hlp->hpp = hpnew;
	return hpnew;
}
 
BLAST_HSPPoolPtr LIBCALL
BlastHSPPoolNew(cnt)
	size_t	cnt;
{
	return (BLAST_HSPPoolPtr)PoolNew(cnt, sizeof(BLAST_HSP), offsetof(BLAST_HSP, next), NULL);
}
 
void LIBCALL
BlastHSPPoolDestruct(hsp_pool)
	BLAST_HSPPoolPtr	hsp_pool;
{
	PoolDestruct((PoolBlkPtr)hsp_pool);
}

BLAST_HSPPtr LIBCALL
BlastHSPGet(hlp)
	BLAST_HitListPtr	hlp;
{
	register BLAST_HSPPtr	hp;

	hp = (BLAST_HSPPtr)PoolGet((PoolBlkPtr)hlp->pdata.hsp_pool);
	if (hp == NULL)
		return NULL;
	hp->hlp = hlp;
	hp->next = NULL;
	hp->vsp = NULL;
	hp->kbp = NULL;
	hp->script = NULL;
	hp->class = 0;
	return hp;
}

void LIBCALL
BlastHSPPut(hp)
	register BLAST_HSPPtr	hp;
{
	register BLAST_HitListPDataPtr	pdp;

	pdp = &hp->hlp->pdata;
	if (hp->vsp != NULL)
		PoolPutLink((PoolBlkPtr)pdp->vscore_pool, hp->vsp);
	PoolPut((PoolBlkPtr)pdp->hsp_pool, (VoidPtr)hp);
}

/* BlastHSPPutLink -- return an entire HSP linked list to the free pool */
void LIBCALL
BlastHSPPutLink(hlp)
	register BLAST_HitListPtr	hlp;
{
	BLAST_HSPPtr	hp0;
	register BLAST_HSPPtr	hp;
	register BLAST_HitListPDataPtr	pdp;

	pdp = &hlp->pdata;
	for (hp = (hp0 = *hlp->hpp); hp != NULL; hp = hp->next) {
		if (hp->vsp != NULL)
			PoolPutLink((PoolBlkPtr)pdp->vscore_pool, hp->vsp);
	}
	PoolPutLink((PoolBlkPtr)pdp->hsp_pool, hp0);
	*hlp->hpp = NULL;
}

void LIBCALL
BlastHSPTruncate(hlp, maxcnt)
	BLAST_HitListPtr	hlp;
	int	maxcnt;
{
	register BLAST_HSPPtr	hp, hplast, hpbest;
	register BLAST_HitListPDataPtr	pdp;
	int		cnt;

	if (hlp == NULL)
		return;

	hplast = NULL;
	for (cnt = 0, hpbest = hp = *hlp->hpp; hp != NULL && cnt++ < maxcnt; hp = hp->next) {
		hplast = hp;
		if (hp->evalue < hpbest->evalue)
			hpbest = hp;
	}
	if (hplast != NULL) {
		pdp = &hlp->pdata;
		for (hp = hplast->next; hp != NULL; hp = hp->next) {
			if (hp->vsp != NULL)
				PoolPutLink((PoolBlkPtr)pdp->vscore_pool, hp->vsp);
		}
		PoolPutLink((PoolBlkPtr)pdp->hsp_pool, hplast->next);
		hplast->next = NULL;
	}
	else
		hpbest = NULL;
	hlp->best_hsp = hpbest;
}


/*
	BlastHSPAppend

	Append hp2 to the linked list pointed to by hp1 (which may be NULL).
	Return value:  the head element in the updated linked list.
*/
BLAST_HSPPtr LIBCALL
BlastHSPAppend(hp1, hp2)
	register BLAST_HSPPtr	hp1, hp2;
{
	if (hp1 == NULL)
		return hp2;

	while (hp1->next != NULL)
		hp1 = hp1->next;

	hp1->next = hp2;
	return hp1;
}

/*
	BlastHSPCopy

	Substitute the values in hp2 for those in hp1, EXCEPT the pointer
	to the next element in the linked list.
*/
void LIBCALL
BlastHSPCopy(hp1, hp2)
	BLAST_HSPPtr	hp1, hp2;
{
	BLAST_HSPPtr	next;

	next = hp1->next;
	*hp1 = *hp2;
	hp1->next = next;
	return;
}


/*
	blast_hsp_count

	Count the number of HSPs in the linked list.
*/
size_t LIBCALL
BlastHSPCount(hp)
	register BLAST_HSPPtr	hp;
{
	register size_t	cnt = 0;

	while (hp != NULL) {
		++cnt;
		hp = hp->next;
	}
	return cnt;
}

BLAST_Error LIBCALL
BlastHSPNumber(hp)
	register BLAST_HSPPtr	hp;
{
	register unsigned	cnt = 0;

	if (hp == NULL)
		return blast_errno = BLAST_ERR_INVAL;

	while (hp != NULL) {
		hp->ordnum = cnt++;
		hp = hp->next;
	}
	return BLAST_ERR_NONE;
}

CharPtr LIBCALL
BlastHSPLabel(hp)
	BLAST_HSPPtr	hp;
{
	static char	buf[20];

	if (BlastHSPLabelPutString(hp, buf) != BLAST_ERR_NONE)
		return NULL;
	return buf;
}

BLAST_Error LIBCALL
BlastHSPLabelPutFile(hp, fp)
	BLAST_HSPPtr	hp;
	FILE	*fp;
{
	register unsigned	num, div;

	if (hp == NULL || fp == NULL)
		return blast_errno = BLAST_ERR_INVAL;

	num = hp->ordnum;
	if (num < 26) {
		fputc('A' + num, fp);
		return BLAST_ERR_NONE;
	}
	div = (num / 26) + 1;
	num %= 26;
	fputc('A' + num, fp);
	fprintf(fp, "%d", div);
	return BLAST_ERR_NONE;
}

BLAST_Error LIBCALL
BlastHSPLabelPutString(hp, buf)
	BLAST_HSPPtr	hp;
	CharPtr	buf;
{
	register unsigned	num, div;

	if (hp == NULL || buf == NULL)
		return blast_errno = BLAST_ERR_INVAL;

	num = hp->ordnum;
	if (num < 26) {
		buf[0] = 'A' + num;
		buf[1] = NULLB;
		return BLAST_ERR_NONE;
	}
	div = (num / 26) + 1;
	num %= 26;
	sprintf(buf, "%c%d", 'A' + num, div);
	return BLAST_ERR_NONE;
}
