/* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================*/
#include <ncbi.h>
#include <blast/blast.h>

BLAST_ResCompPtr
BlastResCompNew(ap)
	BLAST_AlphabetPtr	ap;
{
	BLAST_ResCompPtr	rcp;
	size_t	range;

	if (ap == NULL) {
		blast_errno = BLAST_ERR_INVAL;
		return NULL;
	}
	if (ap->initialized == FALSE) {
		blast_errno = BLAST_ERR_INIT;
		return NULL;
	}

	rcp = (BLAST_ResCompPtr) BlastCalloc(sizeof(*rcp));
	if (rcp == NULL)
		return NULL;

	rcp->ap = ap;

	range = ap->letter_max - ap->letter_min + 1;
	rcp->_comp0 = (unsigned long PNTR) BlastCalloc(sizeof(rcp->comp[0]) * range);
	if (rcp->_comp0 == NULL) {
		BlastResCompDestruct(rcp);
		return NULL;
	}

	rcp->comp = rcp->_comp0 - ap->letter_min;

	return rcp;
}


BLAST_Error
BlastResCompDestruct(rcp)
	BLAST_ResCompPtr	rcp;
{
	if (rcp == NULL)
		return BLAST_ERR_NONE;

	if (rcp->_comp0 != NULL)
		BlastFree(rcp->_comp0);
	Nlm_MemSet((CharPtr)rcp, 0, sizeof(*rcp));
	BlastFree((CharPtr)rcp);
	return BLAST_ERR_NONE;
}

BLAST_Error
BlastResCompCpy(dst, src)
	BLAST_ResCompPtr	dst, src;
{
	size_t	range;

	if (src == NULL || dst == NULL || src->ap == NULL || dst->ap == NULL)
		return blast_errno = BLAST_ERR_INVAL;
	if (src->ap != dst->ap)
		return blast_errno = BLAST_ERR_ALPHACONFLICT;

	range = src->ap->letter_max - src->ap->letter_min + 1;
	Nlm_MemCpy((Nlm_VoidPtr)dst->_comp0, (Nlm_VoidPtr)src->_comp0, sizeof(src->comp[0]) * range);
	return BLAST_ERR_NONE;
}

BLAST_Error
BlastResCompSet(BLAST_ResCompPtr rcp, BLAST_Letter ch, unsigned long cnt)
{
	BLAST_AlphabetPtr	ap;

	ap = rcp->ap;
	if (!BlastAlphabetTst(ap, ch))
		return blast_errno = BLAST_ERR_ALPHAVAL;

	rcp->comp[ch] = cnt;
	return BLAST_ERR_NONE;
}

BLAST_Error
BlastResCompClr(rcp)
	BLAST_ResCompPtr	rcp;
{
	int	i;

	for (i = rcp->ap->letter_min; i <= (int)rcp->ap->letter_max; ++i)
		rcp->comp[i] = 0;

	return BLAST_ERR_NONE;
}

unsigned long
BlastResCompSum(rcp)
	register BLAST_ResCompPtr	rcp;
{
	register unsigned long	sum = 0;
	register int		i;

	if (rcp == NULL) {
		blast_errno = BLAST_ERR_INVAL;
		return 0;
	}

	for (i = rcp->ap->letter_min; i <= (int)rcp->ap->letter_max; ++i)
		sum += rcp->comp[i];
	return sum;
}



BLAST_Error
BlastResCompBuf(BLAST_ResCompPtr rcp, BLAST_LetterPtr lp, size_t len, Nlm_Boolean clr)
{
	register BLAST_LetterPtr	lpmax;
	register unsigned long PNTR	comp;

	if (rcp == NULL || rcp->ap == NULL || lp == NULL)
		return blast_errno = BLAST_ERR_INVAL;

	if (clr)
		BlastResCompClr(rcp);

	for (lpmax = lp + len, comp = rcp->comp; lp < lpmax; ++lp)
		++comp[*lp];

	return BLAST_ERR_NONE;
}


BLAST_Error
BlastResCompBufSubtract(rcp, lp, len)
	BLAST_ResCompPtr	rcp;
	BLAST_LetterPtr	lp;
	size_t			len;
{
	register BLAST_LetterPtr	lpmax;
	register BLAST_Letter	ch;
	register unsigned long PNTR	comp;

	if (rcp == NULL || rcp->ap == NULL || lp == NULL)
		return blast_errno = BLAST_ERR_INVAL;
	for (lpmax = lp + len, comp = rcp->comp; lp < lpmax; ++lp)
		if (comp[ch = *lp] > 0)
			--comp[ch];

	return BLAST_ERR_NONE;
}

BLAST_Error
BlastResCompStr(BLAST_ResCompPtr rcp, BLAST_StrPtr str, Nlm_Boolean clr)
{
	BLAST_LetterPtr	lp, lpmax;

	if (rcp == NULL || str == NULL)
		return blast_errno = BLAST_ERR_INVAL;
	if (rcp->ap != str->ap)
		return blast_errno = BLAST_ERR_ALPHACONFLICT;
	if (str->lpb != 1)
		return blast_errno = BLAST_ERR_ENCODING;

	if (clr)
		BlastResCompClr(rcp);

	for (lp = str->str, lpmax = str->str + str->len; lp < lpmax; ++lp)
		++rcp->comp[*lp];

	return BLAST_ERR_NONE;
}


BLAST_Error
BlastResCompStrSubtract(rcp, str)
	BLAST_ResCompPtr	rcp;
	BLAST_StrPtr	str;
{
	BLAST_LetterPtr	lp, lpmax;

	if (rcp == NULL || str == NULL)
		return blast_errno = BLAST_ERR_INVAL;
	if (rcp->ap != str->ap)
		return blast_errno = BLAST_ERR_ALPHACONFLICT;
	if (str->lpb != 1)
		return blast_errno = BLAST_ERR_ENCODING;

	for (lp = str->str, lpmax = str->str + str->len; lp < lpmax; ++lp)
		if (rcp->comp[*lp] > 0)
			--rcp->comp[*lp];

	return BLAST_ERR_NONE;
}
