/* ===========================================================================
*
*                            PUBLIC DOMAIN NOTICE
*               National Center for Biotechnology Information
*
*  This software/database is a "United States Government Work" under the
*  terms of the United States Copyright Act.  It was written as part of
*  the author's official duties as a United States Government employee and
*  thus cannot be copyrighted.  This software/database is freely available
*  to the public for use. The National Library of Medicine and the U.S.
*  Government have not placed any restriction on its use or reproduction.
*
*  Although all reasonable efforts have been taken to ensure the accuracy
*  and reliability of the software and data, the NLM and the U.S.
*  Government do not and cannot warrant the performance or results that
*  may be obtained by using this software or data. The NLM and the U.S.
*  Government disclaim all warranties, express or implied, including
*  warranties of performance, merchantability or fitness for any particular
*  purpose.
*
*  Please cite the author in any work or product based on this material.
*
* ===========================================================================*/
#include <ncbi.h>
#include <blast/blast.h>

typedef struct {
		BLAST_Letter	ch;
		double			p;
	} BLAST_LetterProb;

static BLAST_ResFreqPtr	master_r;

#if STD_AMINO_ACID_FREQS == Robinson_prob
/*  Robinson, A.B. and L.R. Robinson, PNAS 88:8880-8884 (1991) */
static BLAST_LetterProb	Robinson_prob[] = {
		{ 'A', 35155. },
		{ 'C',  8669. },
		{ 'D', 24161. },
		{ 'E', 28354. },
		{ 'F', 17367. },
		{ 'G', 33229. },
		{ 'H',  9906. },
		{ 'I', 23161. },
		{ 'K', 25872. },
		{ 'L', 40625. },
		{ 'M', 10101. },
		{ 'N', 20212. },
		{ 'P', 23435. },
		{ 'Q', 19208. },
		{ 'R', 23105. },
		{ 'S', 32070. },
		{ 'T', 26311. },
		{ 'V', 29012. },
		{ 'W',  5990. },
		{ 'Y', 14488. }
	};
#endif

#if STD_AMINO_ACID_FREQS == Dayhoff_prob
/*  M. O. Dayhoff amino acid background frequencies   */
static BLAST_LetterProb	Dayhoff_prob[] = {
		{ 'A', 87.13 },
		{ 'C', 33.47 },
		{ 'D', 46.87 },
		{ 'E', 49.53 },
		{ 'F', 39.77 },
		{ 'G', 88.61 },
		{ 'H', 33.62 },
		{ 'I', 36.89 },
		{ 'K', 80.48 },
		{ 'L', 85.36 },
		{ 'M', 14.75 },
		{ 'N', 40.43 },
		{ 'P', 50.68 },
		{ 'Q', 38.26 },
		{ 'R', 40.90 },
		{ 'S', 69.58 },
		{ 'T', 58.54 },
		{ 'V', 64.72 },
		{ 'W', 10.49 },
		{ 'Y', 29.92 }
	};
#endif

#if STD_AMINO_ACID_FREQS == Altschul_prob
/* Stephen Altschul amino acid background frequencies */
static BLAST_LetterProb Altschul_prob[] = {
		{ 'A', 81.00 },
		{ 'C', 15.00 },
		{ 'D', 54.00 },
		{ 'E', 61.00 },
		{ 'F', 40.00 },
		{ 'G', 68.00 },
		{ 'H', 22.00 },
		{ 'I', 57.00 },
		{ 'K', 56.00 },
		{ 'L', 93.00 },
		{ 'M', 25.00 },
		{ 'N', 45.00 },
		{ 'P', 49.00 },
		{ 'Q', 39.00 },
		{ 'R', 57.00 },
		{ 'S', 68.00 },
		{ 'T', 58.00 },
		{ 'V', 67.00 },
		{ 'W', 13.00 },
		{ 'Y', 32.00 }
	};
#endif

static BLAST_LetterProb	nt_prob[] = {
		{ 'A', 25.00 },
		{ 'C', 25.00 },
		{ 'G', 25.00 },
		{ 'T', 25.00 }
	};



BLAST_ResFreqPtr LIBCALL
BlastResFreqNew(ap)
	BLAST_AlphabetPtr	ap;
{
	BLAST_ResFreqPtr	rfp;
	size_t	range;

	if (ap->initialized == FALSE) {
		blast_errno = BLAST_ERR_INIT;
		return NULL;
	}

	rfp = (BLAST_ResFreqPtr) BlastCalloc(sizeof(*rfp));
	if (rfp == NULL)
		return NULL;

	rfp->ap = ap;

	range = ap->letter_max - ap->letter_min + 1;
	rfp->_prob0 = (double PNTR) BlastCalloc(sizeof(rfp->prob[0]) * range);
	if (rfp->_prob0 == NULL) {
		BlastResFreqDestruct(rfp);
		return NULL;
	}
	rfp->prob = rfp->_prob0 - ap->letter_min;

	return rfp;
}

BLAST_Error LIBCALL
BlastResFreqDestruct(rfp)
	BLAST_ResFreqPtr	rfp;
{
	if (rfp == NULL)
		return BLAST_ERR_NONE;

	if (rfp->_prob0 != NULL)
		BlastFree(rfp->_prob0);
	Nlm_MemSet((CharPtr)rfp, 0, sizeof(*rfp));
	BlastFree((CharPtr)rfp);
	return BLAST_ERR_NONE;
}

BLAST_Error LIBCALL
BlastResFreqRegister(rfp)
	BLAST_ResFreqPtr	rfp;
{
	if (BlastResFreqIsRegistered(rfp))
		return BLAST_ERR_NONE;
	if (master_r != NULL)
		master_r->previous = rfp;
	rfp->previous = NULL;
	rfp->next = master_r;
	master_r = rfp;
	return BLAST_ERR_NONE;
}

Nlm_Boolean LIBCALL
BlastResFreqIsRegistered(rfp)
	BLAST_ResFreqPtr	rfp;
{
	BLAST_ResFreqPtr	rfp2;

	for (rfp2 = master_r; rfp2 != NULL; rfp2 = rfp2->next) {
		if (rfp == rfp2)
			return TRUE;
	}
	return FALSE;
}

BLAST_ResFreqPtr LIBCALL
BlastResFreqFind(ap)
	BLAST_AlphabetPtr	ap;
{
	BLAST_ResFreqPtr	rfp;
	BLAST_AlphabetPtr	iupacap;
	BLAST_AlphaMapPtr	amp;
	BLAST_Letter	ch;
	int		i;

	/* See if the object is already registered */
	for (rfp = master_r; rfp != NULL; rfp = rfp->next) {
		if (rfp->ap == ap)
			return rfp;
	}

	/* Create the requested object, if possible */
	if (ap->alphatype == BLAST_ALPHATYPE_AMINO_ACID) {
		iupacap = BlastAlphabetFindByID(BLAST_ALPHA_IUPACaa);
	}
	else
		if (ap->alphatype == BLAST_ALPHATYPE_NUCLEIC_ACID) {
			iupacap = BlastAlphabetFindByID(BLAST_ALPHA_IUPACna);
		}
		else {
			blast_errno = BLAST_ERR_NF;
			return NULL;
		}
	if (iupacap == NULL)
		return NULL;
	if (iupacap == ap)
		amp = NULL; /* no translation/mapping required */
	else {
		amp = BlastAlphaMapFind(iupacap, ap);
		if (amp == NULL)
			return NULL;
	}

	rfp = BlastResFreqNew(ap);
	if (rfp == NULL)
		return NULL;

	if (ap->alphatype == BLAST_ALPHATYPE_AMINO_ACID)
		for (i = 0; i < DIM(STD_AMINO_ACID_FREQS); ++i) {
			ch = STD_AMINO_ACID_FREQS[i].ch;
			if (amp != NULL) {
				if (!BlastAlphaMapTst(amp, ch))
					continue;
				ch = BlastAlphaMapChr(amp, ch);
			}
			rfp->prob[ch] = STD_AMINO_ACID_FREQS[i].p;
		}
	if (ap->alphatype == BLAST_ALPHATYPE_NUCLEIC_ACID)
		for (i = 0; i < DIM(nt_prob); ++i) {
			ch = nt_prob[i].ch;
			if (amp != NULL) {
				if (!BlastAlphaMapTst(amp, ch))
					continue;
				ch = BlastAlphaMapChr(amp, ch);
			}
			rfp->prob[ch] = nt_prob[i].p;
		}

	if (BlastResFreqNormalize(rfp, 1.0) != BLAST_ERR_NONE)
		return NULL;
	if (BlastResFreqRegister(rfp) != BLAST_ERR_NONE)
		return NULL;
	return rfp;
}

BLAST_ResFreqPtr LIBCALL
BlastResFreqStr(str)
	BLAST_StrPtr	str;
{
	BLAST_ResCompPtr	rcp = NULL;
	BLAST_ResFreqPtr	rfp = NULL;

	if (str == NULL) {
		blast_errno = BLAST_ERR_INVAL;
		return NULL;
	}
	if (str->lpb != 1) {
		blast_errno = BLAST_ERR_ENCODING;
		return NULL;
	}

	if ((rfp = BlastResFreqNew(str->ap)) == NULL)
		goto Error;
	if ((rcp = BlastResCompNew(str->ap)) == NULL)
		goto Error;

	if (BlastResCompStr(rcp, str, FALSE) != BLAST_ERR_NONE)
		goto Error;

	if (BlastResFreqResComp(rfp, rcp) != BLAST_ERR_NONE)
		goto Error;

	BlastResCompDestruct(rcp);
	return rfp;

Error:
	if (rcp != NULL)
		BlastResCompDestruct(rcp);
	if (rfp != NULL)
		BlastResFreqDestruct(rfp);
	return NULL;
}

BLAST_Error LIBCALL
BlastResFreqNormalize(rfp, norm)
	BLAST_ResFreqPtr	rfp;
	double	norm;
{
	BLAST_AlphabetPtr	ap = rfp->ap;
	int		i;
	double	sum = 0., p;

	if (norm == 0.)
		return blast_errno = BLAST_ERR_DOMAIN;

	for (i=0; i<ap->alphasize; ++i) {
		p = rfp->prob[ap->alist[i]];
		if (p < 0.)
			return blast_errno = BLAST_ERR_RESFREQ;
		sum += p;
	}
	if (sum <= 0.)
		return BLAST_ERR_NONE;
	for (i=0; i<ap->alphasize; ++i) {
		rfp->prob[ap->alist[i]] /= sum;
		rfp->prob[ap->alist[i]] *= norm;
	}
	return BLAST_ERR_NONE;
}

BLAST_Error LIBCALL
BlastResFreqSet(BLAST_ResFreqPtr rfp, BLAST_Letter ch, double p)
{
	BLAST_AlphabetPtr	ap;

	ap = rfp->ap;
	if (!BlastAlphabetTst(ap, ch))
		return blast_errno = BLAST_ERR_ALPHAVAL;

	rfp->prob[ch] = p;
	return BLAST_ERR_NONE;
}

BLAST_Error LIBCALL
BlastResFreqClr(rfp)
	BLAST_ResFreqPtr	rfp;
{
	long	i;

	for (i=rfp->ap->letter_min; i <= (int)rfp->ap->letter_max; ++i)
		rfp->prob[i] = 0.;

	return BLAST_ERR_NONE;
}

BLAST_Error LIBCALL
BlastResFreqBuf(rfp, lp, len)
	BLAST_ResFreqPtr	rfp;
	BLAST_LetterPtr	lp;
	size_t	len;
{
	size_t	cntbuf[BLAST_LETTER_MAX - BLAST_LETTER_MIN + 1], PNTR cnt;
	BLAST_LetterPtr	lpmax;
	int		i;

	cnt = cntbuf - rfp->ap->letter_min;
	for (i = rfp->ap->letter_min; i <= (int)rfp->ap->letter_max; ++i)
		cnt[i] = 0;

	lpmax = lp + len;
	while (lp < lpmax)
		++cnt[*lp++];

	for (i = rfp->ap->letter_min; i <= (int)rfp->ap->letter_max; ++i)
		rfp->prob[i] = cnt[i];

	return BLAST_ERR_NONE;
}

/*
	Calculate the residue frequencies associated with the provided ResComp
*/
BLAST_Error LIBCALL
BlastResFreqResComp(rfp, rcp)
	BLAST_ResFreqPtr	rfp;
	BLAST_ResCompPtr	rcp;
{
	int		i;
	double	sum = 0.;

	if (rfp == NULL || rcp == NULL)
		return blast_errno = BLAST_ERR_INVAL;

	if (rfp->ap != rcp->ap)
		return blast_errno = BLAST_ERR_ALPHACONFLICT;

	for (i = rcp->ap->letter_min; i <= (int)rcp->ap->letter_max; ++i)
		sum += rcp->comp[i];

	if (sum == 0.) {
		BlastResFreqClr(rfp);
		return BLAST_ERR_NONE;
	}

	for (i = rcp->ap->letter_min; i <= (int)rcp->ap->letter_max; ++i)
		rfp->prob[i] = rcp->comp[i] / sum;

	return BLAST_ERR_NONE;
}

double LIBCALL
BlastResFreqStrAlist(str, ap, alist, alistlen)
	BLAST_StrPtr	str;
	BLAST_AlphabetPtr	ap;
	BLAST_LetterPtr	alist;
	size_t	alistlen;
{
	BLAST_ResFreqPtr	rfp;
	BLAST_AlphaMapPtr	amp;
	BLAST_Letter	ch;
	BLAST_LetterPtr	alistmax;
	double	sum = 0.;

	if (str == NULL || ap == NULL || alist == NULL)
		return -1.;

	rfp = BlastResFreqStr(str);
	if (rfp == NULL)
		return -1.;

	alistmax = alist + alistlen;

	if (str->ap != ap) {
		if ((amp = BlastAlphaMapFind(ap, str->ap)) == NULL)
			return -1.;
		for (; alist < alistmax; ++alist) {
			if (amp->maptst[ch = *alist]) {
				sum += rfp->prob[amp->map[ch]];
			}
		}
	}
	else
		for (; alist < alistmax; ++alist) {
			sum += rfp->prob[*alist];
		}

	BlastResFreqDestruct(rfp);
	return sum;
}
